import warnings

try:
    from .postgres import PostgresDatabase as PostgresDatabase
except ModuleNotFoundError:
    warnings.warn("PostgresDatabase will not be available", ImportWarning)

from .sqlite import SQLiteDatabase as SQLiteDatabase



import logging
import textwrap
import typing

import psycopg2
import psycopg2.extras
import psycopg2.pool


class PostgresDatabase:
    def __init__(self, dsn: str, minconn: int = 1, maxconn: int = 1) -> None:
        self.log = logging.getLogger(__name__)
        self.p = psycopg2.pool.ThreadedConnectionPool(
            minconn, maxconn, dsn, cursor_factory=psycopg2.extras.DictCursor
        )
        psycopg2.extras.register_uuid()

    def b(self, sql: str, records: list[dict]) -> None:
        """Batch execute a query"""
        cnx = self.p.getconn()
        try:
            with cnx:
                with cnx.cursor() as cur:
                    self.log.debug(f"Batch query with {len(records)} parameter sets")
                    self.log.debug(textwrap.dedent(sql))
                    psycopg2.extras.execute_batch(cur, sql, records)
        finally:
            self.p.putconn(cnx)

    def q(self, sql: str, params: dict | None = None) -> list[dict]:
        """Execute a query and return all rows"""
        if params is None:
            params = {}
        cnx = self.p.getconn()
        try:
            with cnx:
                with cnx.cursor() as c:
                    self.log.debug(textwrap.dedent(c.mogrify(sql, params).decode()))
                    c.execute(sql, params)
                    return c.fetchall()
        finally:
            self.p.putconn(cnx)

    def q_one(self, sql: str, params: dict | None = None) -> dict | None:
        """Execute a query and return the first row, or None if there are no rows"""
        for r in self.q(sql, params):
            return r
        return None

    def q_val(self, sql: str, params: dict | None = None) -> typing.Any:  # noqa: ANN401
        """Execute a query and return the value in the first column of the first row,
        or None if there are no rows"""
        for r in self.q(sql, params):
            return r[0]
        return None

    def u(self, sql: str, params: dict | None = None) -> int:
        """Execute a statement and return the number of rows affected"""
        if params is None:
            params = {}
        cnx = self.p.getconn()
        try:
            with cnx:
                with cnx.cursor() as c:
                    self.log.debug(textwrap.dedent(c.mogrify(sql, params).decode()))
                    c.execute(sql, params)
                    return c.rowcount
        finally:
            self.p.putconn(cnx)



import datetime
import decimal
import logging
import sqlite3
import typing
import uuid


def register_adapters_and_converters() -> None:
    def convert_bool(value: bytes) -> bool:
        return value == b"True"

    sqlite3.register_adapter(bool, str)
    sqlite3.register_converter("bool", convert_bool)

    def convert_date(value: bytes) -> datetime.date:
        return datetime.date.fromisoformat(value.decode())

    sqlite3.register_adapter(datetime.date, str)
    sqlite3.register_converter("date", convert_date)

    def convert_decimal(value: bytes) -> decimal.Decimal:
        return decimal.Decimal(value.decode())

    sqlite3.register_adapter(decimal.Decimal, str)
    sqlite3.register_converter("decimal", convert_decimal)

    def convert_uuid(value: bytes) -> uuid.UUID:
        return uuid.UUID(value.decode())

    sqlite3.register_adapter(uuid.UUID, str)
    sqlite3.register_converter("uuid", convert_uuid)


register_adapters_and_converters()


class SQLiteDatabase:
    def __init__(self, dsn: str) -> None:
        self.log = logging.getLogger(__name__)
        self.cnx = sqlite3.connect(dsn, detect_types=sqlite3.PARSE_DECLTYPES)
        self.cnx.isolation_level = None
        self.cnx.row_factory = sqlite3.Row
        self.cnx.set_trace_callback(self.log.debug)

    def _q_gen(self, sql: str, params: dict | None = None) -> typing.Iterator[dict]:
        if params is None:
            params = {}
        yield from self.cnx.execute(sql, params)

    def b(self, sql: str, params: list[dict]) -> None:
        self.cnx.executemany(sql, params)

    def q(self, sql: str, params: dict | None = None) -> list[dict]:
        return list(self._q_gen(sql, params))

    def q_one(self, sql: str, params: dict | None = None) -> dict | None:
        for r in self._q_gen(sql, params):
            return r
        return None

    def q_val(self, sql: str, params: dict | None = None) -> typing.Any:  # noqa: ANN401
        for r in self._q_gen(sql, params):
            return r[0]
        return None

    def u(self, sql: str, params: dict | None = None) -> int:
        if params is None:
            params = {}
        c = self.cnx.execute(sql, params)
        return c.rowcount
